#!/bin/python3
import serial
import ctypes as ct
import signal
import sys
import time
import os 
import subprocess


VOXL_PLATFORM = False
try:
	ret = subprocess.check_output(['voxl-platform'])
	VOXL_PLATFORM = True
except FileNotFoundError:
	VOXL_PLATFORM = False

# Detect if we are on a voxl2 or host computer
if VOXL_PLATFORM:
    import python_packages.voxl_utils_python.voxl_serial as voxl_serial
    BASE_DIR = '/usr/share/modalai/voxl-elrs'
    host = "voxl2"
else:
    BASE_DIR = '/usr/share/modalai/voxl-elrs'
    host = "ubuntu"

LOG_DIR = f"{BASE_DIR}/logs"
os.makedirs(LOG_DIR, exist_ok=True)
LOG_FILE=f"{LOG_DIR}/M0139_RX.log"
# LOG_FILE_RAW=f"{LOG_DIR}/M0139_RX_RAW.log"

uint8_t = ct.c_uint8
if host == "voxl2":
    dev = voxl_serial.VoxlSerialPort()
    dev.port = '/dev/tty-slpi-7'
    dev.baudrate = 420000
    dev.open()
else:
    # Automatically opens serial
    print("Using /dev/ttyUSB0")
    dev = serial.Serial("/dev/ttyUSB0",115200, timeout=0)

def sigint_handler(signal, frame):
    print("Exiting gracefully...")
    dev.close()  # Close the serial port before exiting
    sys.exit(0)

# Register the SIGINT handler
signal.signal(signal.SIGINT, sigint_handler)

# with open(LOG_FILE,'w') as log, open(LOG_FILE_RAW, 'wb') as log_raw:
with open(LOG_FILE,'w') as log:
    while(True):
        if (dev.inWaiting()>0):
            data = dev.read(dev.inWaiting())
            # Print all values
            print(data)
            msg = "".join([chr(value) for value in data if 0 <= value <= 127])
            print(f"\nDecoded {len(data)} bytes:\n{msg}")
            log.write(msg)
            log.flush()
            # log_raw.write(data)
            # time.sleep(0.25)
